# -*- coding: utf-8 -*-
# Copyright (c) 2015 Spotify AB
from __future__ import absolute_import, division, print_function

import os

import pytest
import xmltodict

from ramlfications import parser as pw
from ramlfications.config import setup_config
from ramlfications.raml import RootNode, ResourceTypeNode, TraitNode
from ramlfications._helpers import load_file

from .base import EXAMPLES


@pytest.fixture(scope="session")
def loaded_raml():
    raml_file = os.path.join(EXAMPLES + "complete-valid-example.raml")
    return load_file(raml_file)


@pytest.fixture(scope="session")
def root():
    raml_file = os.path.join(EXAMPLES + "complete-valid-example.raml")
    loaded_raml_file = load_file(raml_file)
    config = setup_config(EXAMPLES + "test-config.ini")
    return pw.create_root(loaded_raml_file, config)


def test_parse_raml(loaded_raml):
    config = setup_config(EXAMPLES + "test-config.ini")
    root = pw.parse_raml(loaded_raml, config)
    assert isinstance(root, RootNode)


def test_create_root(root):
    assert isinstance(root, RootNode)


def test_base_uri(root):
    expected = "https://{subdomain}.example.com/v1/{communityPath}"
    assert expected == root.base_uri


def test_protocols(root):
    expected = ["HTTPS"]
    assert expected == root.protocols


def test_docs(root):
    exp_title = "Example Web API Docs"
    exp_content = ("Welcome to the _Example Web API_ demo specification. "
                   "This is *not* the complete API\nspecification, and is "
                   "meant for testing purposes within this RAML "
                   "specification.\n")

    assert exp_title == root.documentation[0].title.raw
    assert exp_title == repr(root.documentation[0].title)
    assert exp_content == root.documentation[0].content.raw
    assert exp_content == repr(root.documentation[0].content)

    title_html = "<p>Example Web API Docs</p>\n"
    content_html = ("<p>Welcome to the <em>Example Web API</em> demo "
                    "specification. This is <em>not</em> the complete API\n"
                    "specification, and is meant for testing purposes within "
                    "this RAML specification.</p>\n")
    assert title_html == root.documentation[0].title.html
    assert content_html == root.documentation[0].content.html


def test_base_uri_params(root):
    exp_name = "subdomain"
    exp_desc = "subdomain of API"
    exp_desc_html = "<p>subdomain of API</p>\n"
    exp_default = "api"

    assert exp_name == root.base_uri_params[0].name
    assert exp_desc == root.base_uri_params[0].description.raw
    assert exp_desc_html == root.base_uri_params[0].description.html
    assert exp_default == root.base_uri_params[0].default


def test_uri_params(root):
    assert root.uri_params[0].name == "communityPath"
    assert root.uri_params[0].display_name == "Community Path"
    assert root.uri_params[0].type == "string"
    assert root.uri_params[0].min_length == 1
    assert root.uri_params[0].description is None
    assert not hasattr(root.uri_params[0].description, "raw")
    assert not hasattr(root.uri_params[0].description, "html")
    assert root.uri_params[0].default is None
    assert root.uri_params[0].enum is None
    assert root.uri_params[0].example is None


def test_title(root):
    exp_name = "Example Web API"
    assert exp_name == root.title


def test_version(root):
    exp_version = "v1"
    assert exp_version == root.version


def test_schemas(root):
    thingy_json_schema = {'Thingy': {'name': 'New Thingy', 'public': False}}
    assert thingy_json_schema == root.schemas[0]

    thingy_xsd_schema = {
        'ThingyXsd': {
            'xs:schema': {
                '@attributeFormDefault': 'unqualified',
                '@elementFormDefault': 'qualified',
                '@xmlns:xs': 'http://www.w3.org/2001/XMLSchema',
                'xs:element': {
                    '@name': 'thingy',
                    '@type': 'thingyType'
                },
                'xs:complexType': {
                    '@name': 'thingyType',
                    'xs:sequence': {
                        'xs:element': {
                            '@type': 'xs:string',
                            '@name': 'name',
                            '@minOccurs': '1',
                            '@maxOccurs': '1'
                        }
                    }
                }
            }
        }
    }
    assert thingy_xsd_schema == root.schemas[1]

    thingy_xsd_list_schema = {
        'ThingyListXsd': {
            'xs:schema': {
                '@attributeFormDefault': 'unqualified',
                '@elementFormDefault': 'qualified',
                '@xmlns:xs': 'http://www.w3.org/2001/XMLSchema',
                'xs:include': {
                    '@schemaLocation': './thingy.xsd'
                },
                'xs:element': {
                    '@name': 'thingies',
                    'xs:complexType': {
                        'xs:sequence': {
                            '@minOccurs': '0',
                            '@maxOccurs': 'unbounded',
                            'xs:element': {
                                '@type': 'thingyType',
                                '@name': 'thingy'
                            }
                        }
                    }
                }
            }
        }
    }
    assert thingy_xsd_list_schema == root.schemas[2]


def test_media_type(root):
    exp_media_type = "application/json"
    assert exp_media_type == root.media_type


@pytest.fixture(scope="session")
def api():
    raml_file = os.path.join(EXAMPLES + "complete-valid-example.raml")
    loaded_raml_file = load_file(raml_file)
    config = setup_config(EXAMPLES + "test-config.ini")
    return pw.parse_raml(loaded_raml_file, config)


#####
# Test Security Schemes
#####
@pytest.fixture(scope="session")
def sec_schemes():
    raml_file = os.path.join(EXAMPLES + "complete-valid-example.raml")
    loaded_raml_file = load_file(raml_file)
    config = setup_config(EXAMPLES + "test-config.ini")
    api = pw.parse_raml(loaded_raml_file, config)
    return api.security_schemes


def test_create_security_schemes(sec_schemes):
    assert len(sec_schemes) == 3
    assert sec_schemes[0].name == "oauth_2_0"
    assert sec_schemes[0].type == "OAuth 2.0"

    desc = ("Example API supports OAuth 2.0 for authenticating all API "
            "requests.\n")
    assert sec_schemes[0].description.raw == desc

    assert len(sec_schemes[0].headers) == 2
    assert sec_schemes[0].headers[0].name == "Authorization"
    assert sec_schemes[0].headers[1].name == "X-Foo-Header"

    assert len(sec_schemes[0].responses) == 2
    assert sec_schemes[0].responses[0].code == 401

    desc = ("Bad or expired token. This can happen if the user revoked a "
            "token or\nthe access token has expired. You should "
            "re-authenticate the user.\n")
    assert sec_schemes[0].responses[0].description.raw == desc

    assert sec_schemes[0].responses[1].code == 403

    desc = ("Bad OAuth request (wrong consumer key, bad nonce, expired\n"
            "timestamp...). Unfortunately, re-authenticating the user won't "
            "help here.\n")
    assert sec_schemes[0].responses[1].description.raw == desc

    settings = {
        "authorizationUri": "https://accounts.example.com/authorize",
        "accessTokenUri": "https://accounts.example.com/api/token",
        "authorizationGrants": ["code", "token"],
        "scopes": [
            "user-public-profile",
            "user-email",
            "user-activity",
            "nsa-level-privacy"
        ]
    }
    assert sec_schemes[0].settings == settings


def test_create_security_schemes_custom(sec_schemes):
    custom = sec_schemes[1]

    assert custom.name == "custom_auth"
    assert custom.type == "Custom Auth"
    assert len(custom.uri_params) == 1
    assert len(custom.form_params) == 1
    assert len(custom.query_params) == 1
    assert len(custom.body) == 1

    assert custom.uri_params[0].name == "subDomain"
    assert custom.query_params[0].name == "fooQParam"
    assert custom.form_params[0].name == "fooFormParam"
    assert custom.body[0].mime_type == "application/x-www-form-urlencoded"

    assert custom.documentation[0].title.raw == "foo docs"
    assert custom.documentation[0].content.raw == "foo content"


def test_security_scheme_no_desc(sec_schemes):
    no_desc = sec_schemes[2]

    assert no_desc.name == "no_desc"
    assert no_desc.description is None


#####
# Test Traits
#####
@pytest.fixture(scope="session")
def traits():
    raml_file = os.path.join(EXAMPLES + "complete-valid-example.raml")
    loaded_raml_file = load_file(raml_file)
    config = setup_config(EXAMPLES + "test-config.ini")
    api = pw.parse_raml(loaded_raml_file, config)
    return api.traits


def test_create_traits(api):
    for trait in api.traits:
        assert isinstance(trait, TraitNode)


def test_trait_query_params(traits):
    trait = traits[0]
    assert trait.name == "filterable"
    assert trait.usage == "Some description about using filterable"
    assert trait.query_params[0].name == "fields"
    assert trait.query_params[0].type == "string"
    assert trait.query_params[0].display_name == "Fields"

    example = "gizmos.items(added_by.id,gizmo(name,href,widget(name,href)))"
    assert trait.query_params[0].example == example


def test_trait_headers(traits):
    trait = traits[0]
    assert trait.headers[0].name == "X-example-header"
    assert trait.headers[0].description.raw == "An example of a trait header"
    html_desc = "<p>An example of a trait header</p>\n"
    assert trait.headers[0].description.html == html_desc


def test_trait_body(traits):
    trait = traits[0]
    assert trait.body[0].mime_type == "application/json"
    assert trait.body[0].schema == {"name": "string"}
    assert trait.body[0].example == {"name": "example body for trait"}


def test_trait_uri_params(traits):
    trait = traits[4]
    assert trait.uri_params[0].name == "communityPath"
    assert trait.uri_params[0].type == "string"
    assert trait.uri_params[0].example == 'baz-community'

    desc = "The community path URI params trait"
    assert trait.uri_params[0].description.raw == desc
    assert repr(trait.uri_params[0].description) == desc

    desc_html = "<p>The community path URI params trait</p>\n"
    assert trait.uri_params[0].description.html == desc_html


def test_trait_form_params(traits):
    trait = traits[2]
    assert trait.form_params[0].name == "foo"
    assert trait.form_params[0].display_name == "Foo"
    assert trait.form_params[0].type == "string"
    assert trait.form_params[0].default == "bar"
    assert trait.form_params[0].min_length == 5
    assert trait.form_params[0].max_length == 50
    assert trait.form_params[0].description.raw == "The Foo Form Field"

    trait_desc = "A description of a trait with form parameters"
    assert trait.description.raw == trait_desc
    media_type = "application/x-www-form-urlencoded"
    assert trait.media_type == media_type


def test_trait_base_uri_params(traits):
    trait = traits[3]
    assert trait.base_uri_params[0].name == "communityPath"
    assert trait.base_uri_params[0].display_name == "Community Path trait"
    assert trait.base_uri_params[0].type == "string"
    assert trait.base_uri_params[0].example == "baz-community"

    param_desc = "The community path base URI trait"
    assert trait.base_uri_params[0].description.raw == param_desc

    trait_desc = "A description of a trait with base URI parameters"
    assert trait.description.raw == trait_desc


#####
# Test Resource Types
#####
@pytest.fixture(scope="session")
def resource_types():
    raml_file = os.path.join(EXAMPLES + "complete-valid-example.raml")
    loaded_raml_file = load_file(raml_file)
    config = setup_config(EXAMPLES + "test-config.ini")
    api = pw.parse_raml(loaded_raml_file, config)
    return api.resource_types


def test_create_resource_types(api):
    for res_type in api.resource_types:
        assert isinstance(res_type, ResourceTypeNode)


def test_resource_type(resource_types):
    resource_type = resource_types[0]
    assert resource_type.name == "base"
    assert resource_type.method == "get"
    assert resource_type.optional
    assert resource_type.is_ is None
    assert resource_type.traits is None

    header = resource_type.headers[0]
    assert header.name == "Accept"
    assert header.method == "get"
    assert header.type == "string"
    assert header.description.raw == "Is used to set specified media type."
    assert header.example is None
    assert header.default is None
    assert header.required is False

    body = resource_type.body[0]
    assert body.mime_type == "application/json"
    assert body.schema == {"name": "string"}
    assert body.example == {"name": "Foo Bar"}
    assert body.form_params is None

    response = resource_type.responses[0]
    assert response.code == 403

    desc = ("API rate limit exceeded.\n")
    assert response.description.raw == desc
    assert response.method == "get"

    resp_header = response.headers[0]
    assert resp_header.name == "X-waiting-period"

    resp_desc = ("The number of seconds to wait before you can attempt to "
                 "make a request again.\n")
    assert resp_header.description.raw == resp_desc
    assert resp_header.type == "integer"
    assert resp_header.required

    # TODO: add types to Header object
    assert resp_header.minimum == 1
    assert resp_header.maximum == 3600
    assert resp_header.example == 34

    resp_body = response.body[0]
    assert resp_body.mime_type == "application/json"
    assert resp_body.schema == {"name": "string"}
    assert resp_body.example == {"name": "Foo Bar"}
    assert resp_body.form_params is None


def test_resource_type_method_protocol(resource_types):
    resource = resource_types[-4]
    assert resource.name == "protocolExampleType"
    assert resource.protocols == ["HTTP"]


def test_resource_type_uri_params(resource_types):
    uri_param = resource_types[0].uri_params[0]
    assert uri_param.name == "mediaTypeExtension"

    desc = "Use .json to specify application/json media type."
    assert uri_param.description.raw == desc
    assert uri_param.enum == [".json"]
    assert uri_param.min_length is None
    assert uri_param.type == "string"
    assert uri_param.required
    assert uri_param.default is None


def test_resource_type_query_params(resource_types):
    res = resource_types[3]
    assert res.name == "queryParamType"
    query_param = res.query_params[0]
    assert query_param.name == "ids"
    assert query_param.description.raw == "A comma-separated list of IDs"
    assert query_param.display_name == "Some sort of IDs"
    assert query_param.type == "string"
    assert query_param.required


def test_resource_type_form_params(resource_types):
    res = resource_types[4]
    assert res.name == "formParamType"
    form_param = res.form_params[0]
    assert form_param.name == "aFormParam"
    assert form_param.description.raw == "An uncreative form parameter"
    assert form_param.display_name == "Some sort of Form Parameter"
    assert form_param.type == "string"
    assert form_param.required


def test_resource_type_base_uri_params(resource_types):
    res = resource_types[5]
    assert res.name == "baseUriType"
    base_uri_params = res.base_uri_params[0]
    assert base_uri_params.name == "subdomain"

    desc = "subdomain for the baseUriType resource type"
    assert base_uri_params.description.raw == desc

    assert base_uri_params.default == "fooBar"


def test_resource_type_properties(resource_types):
    another_example = resource_types[6]
    assert another_example.name == "anotherExample"

    desc = "Another Resource Type example"
    assert another_example.description.raw == desc

    usage = "Some sort of usage description"
    assert another_example.usage == usage

    assert another_example.optional is False
    assert another_example.media_type == "text/xml"


def test_resource_type_inherited(resource_types):
    inherited = resource_types[9]
    assert inherited.type == "base"
    assert inherited.usage == "Some sort of usage text"
    assert inherited.display_name == "inherited example"

    inherited_response = inherited.responses[1]
    assert inherited_response.code == 403

    new_response = inherited.responses[2]
    assert new_response.code == 500


def test_resource_type_with_trait(resource_types):
    another_example = resource_types[6]
    assert another_example.is_ == ["filterable"]

    trait = another_example.traits[0]
    assert trait.name == "filterable"

    query_param = trait.query_params[0]
    assert query_param.name == "fields"
    assert query_param.display_name == "Fields"
    assert query_param.type == "string"

    desc = "A comma-separated list of fields to filter query"
    assert query_param.description.raw == desc

    example = "gizmos.items(added_by.id,gizmo(name,href,widget(name,href)))"
    assert query_param.example == example


def test_resource_type_secured_by(resource_types):
    another_example = resource_types[6]
    assert another_example.secured_by == ["oauth_2_0"]

    scheme = another_example.security_schemes[0]
    assert scheme.name == "oauth_2_0"
    assert scheme.type == "OAuth 2.0"

    desc = ("Example API supports OAuth 2.0 for authenticating all API "
            "requests.\n")
    assert scheme.description.raw == desc

    desc_by = {
        "headers": {
            "Authorization": {
                "description": "Used to send a valid OAuth 2 access token.\n",
                "type": "string"
            },
            "X-Foo-Header": {
                "description": "a foo header",
                "type": "string"
            }
        },
        "responses": {
            401: {
                "description": ("Bad or expired token. This can happen if the "
                                "user revoked a token or\nthe access token "
                                "has expired. You should re-authenticate the "
                                "user.\n")
            },
            403: {
                "description": ("Bad OAuth request (wrong consumer key, bad "
                                "nonce, expired\ntimestamp...). Unfortunately,"
                                " re-authenticating the user won't help "
                                "here.\n")
            }
        }
    }
    assert scheme.described_by == desc_by

    settings = {
        "authorizationUri": "https://accounts.example.com/authorize",
        "accessTokenUri": "https://accounts.example.com/api/token",
        "authorizationGrants": ["code", "token"],
        "scopes": [
            "user-public-profile",
            "user-email",
            "user-activity",
            "nsa-level-privacy"
        ]
    }
    assert scheme.settings == settings


def test_resource_type_empty_mapping():
    raml_file = os.path.join(EXAMPLES + "empty-mapping-resource-type.raml")
    loaded_raml_file = load_file(raml_file)
    config = setup_config(EXAMPLES + "test-config.ini")
    config['validate'] = False
    api = pw.parse_raml(loaded_raml_file, config)

    assert len(api.resource_types) == 1

    res = api.resource_types[0]

    assert res.name == "emptyType"
    assert res.raw == {}


def test_resource_type_empty_mapping_headers():
    raml_file = os.path.join(EXAMPLES + "empty-mapping.raml")
    loaded_raml_file = load_file(raml_file)
    config = setup_config(EXAMPLES + "test-config.ini")
    config['validate'] = False
    api = pw.parse_raml(loaded_raml_file, config)

    base_res_type = api.resource_types[0]

    assert len(base_res_type.headers) == 3
    assert base_res_type.headers[-1].description is None


#####
# Test Resources
#####
@pytest.fixture(scope="session")
def resources():
    raml_file = os.path.join(EXAMPLES + "complete-valid-example.raml")
    loaded_raml_file = load_file(raml_file)
    config = setup_config(EXAMPLES + "test-config.ini")
    api = pw.parse_raml(loaded_raml_file, config)
    return api.resources


def test_resource_properties(resources):
    assert resources[0].name == "/widgets"
    assert resources[0].display_name == "several-widgets"
    assert resources[0].method == "get"
    assert resources[0].protocols == ["HTTPS"]
    assert resources[0].media_type == "application/xml"

    assert resources[1].parent.name == "/widgets"
    assert resources[1].path == "/widgets/{id}"

    abs_uri = "http://{subdomain}.example.com/v1/{communityPath}/widgets/{id}"
    assert resources[1].absolute_uri == abs_uri
    assert resources[1].media_type == "application/xml"
    assert resources[1].protocols == ["HTTP"]

    assert resources[2].is_ == ["paged"]
    # assert resources[2].media_type == "application/xml"
    assert resources[12].type == "collection"

    assert resources[3].media_type == "text/xml"


def test_resource_no_method_properties(resources):
    assert resources[-2].method is None
    assert resources[-2].name == "/no_method_properties"

    assert resources[-1].parent.name == resources[-2].name
    assert resources[-1].method == "get"


def test_resource_headers(resources):
    assert resources[0].headers[0].name == "X-bogus-header"
    desc = "just an extra header for funsies"
    assert resources[0].headers[0].description.raw == desc


def test_resource_inherited_properties(resources):
    res = resources[9]
    # actual object data will not match, just names
    res_uri = res.base_uri_params[0].name
    restype_uri = res.resource_type.base_uri_params[0].name
    assert res_uri  == restype_uri

    res = resources[-6]
    assert res.form_params[0] == res.resource_type.form_params[0]

    res = resources[11]
    assert res.is_ == ["protocolTrait"]
    assert res.traits[0].description.raw == "A trait to assign a protocol"
    assert res.traits[0].protocols == ["HTTP"]


def test_resource_assigned_type(resources):
    res = resources[19]

    assert res.display_name == "thingy"
    assert res.method == "get"
    assert res.type == "item"

    res_type_uri = [r.name for r in res.resource_type.uri_params]
    res_uri = [r.name for r in res.uri_params]

    exp_res_type_uri = ["mediaTypeExtension"]
    exp_res_uri = ["communityPath", "user_id", "thingy_id"]

    assert res_type_uri == exp_res_type_uri
    assert res_uri == exp_res_uri

    assert res.headers[0] == res.resource_type.headers[0]
    assert res.body[0] == res.resource_type.body[0]
    assert res.responses[0] == res.resource_type.responses[0]
    assert len(res.headers) == 3
    assert res.headers[0].name == "X-another-header"
    assert res.headers[1].name == "Accept"
    assert res.headers[2].name == "X-example-header"

    res = resources[18]
    assert res.type == "collection"
    assert res.method == "post"
    assert res.form_params[0] == res.resource_type.form_params[0]

    res = resources[11]
    assert res.type == "queryParamType"
    assert res.method == "get"
    assert res.resource_type.query_params[0] == res.query_params[0]

    res = resources[9]
    assert res.type == "baseUriType"
    assert res.method == "get"
    # actual object data will not match, just name
    res_uri = res.base_uri_params[0].name
    restype_uri = res.resource_type.base_uri_params[0].name
    assert res_uri == restype_uri

    res = resources[1]
    assert res.type == "protocolExampleType"
    assert res.resource_type.name == "protocolExampleType"
    assert res.protocols == res.resource_type.protocols


def test_resource_assigned_trait(resources):
    res = resources[10]

    assert res.name == "/search"
    assert res.is_ == ["paged"]
    assert res.traits[0].description.raw == "A description of the paged trait"
    assert res.traits[0].media_type == "application/xml"

    params = [p.name for p in res.query_params]
    t_params = [p.name for p in res.traits[0].query_params]

    for t in t_params:
        assert t in params


def test_resource_protocols(resources):
    res = resources[13]

    assert res.method == "put"
    assert res.name == "/widgets"

    assert res.protocols == ["HTTP"]

    res = resources[16]

    assert res.method == "get"
    assert res.name == "/users/{user_id}"
    assert res.protocols == ["HTTP"]


def test_resource_responses(resources):
    res = resources[17]

    assert res.responses[0].code == 200
    assert res.responses[0].body[0].mime_type == "application/json"
    assert res.responses[0].body[1].mime_type == "text/xml"
    assert len(res.responses[0].body) == 2

    json_schema = {
        "$schema": "http://json-schema.org/draft-03/schema",
        "type": "array",
        "items": {
            "$ref": "schemas/thingy.json"
        }
    }
    assert res.responses[0].body[0].schema == json_schema
    assert res.responses[0].body[1].schema == 'ThingyListXsd'

    res = resources[10]
    headers = [h.name for h in res.responses[0].headers]
    assert sorted(["X-search-header", "X-another-header"]) == sorted(headers)

    schema = res.responses[0].body[0].schema
    assert schema == {"name": "the search body"}

    schema = res.responses[0].body[1].schema
    assert schema == {}
    example = res.responses[0].body[1].example
    assert example == {}

    res = resources[19]
    headers = [h.name for h in res.responses[0].headers]
    assert "X-waiting-period" in headers
    body = res.responses[0].body[0].schema
    assert body == {"name": "string"}
    codes = [r.code for r in res.responses]
    assert [200, 403] == sorted(codes)

    res = resources[-9]

    assert res.path == "/users/{user_id}/thingys/{thingy_id}"
    assert res.type == "item"
    assert res.responses[0] == res.resource_type.responses[0]
    assert len(res.responses[0].headers) == 1
    assert res.responses[0].headers[0].name == "X-waiting-period"
    assert res.responses[0].headers[0].type == "integer"
    assert res.responses[0].headers[0].minimum == 1
    assert res.responses[0].headers[0].maximum == 3600
    assert res.responses[0].headers[0].example == 34

    desc = ("The number of seconds to wait before you can attempt to make "
            "a request again.\n")
    assert res.responses[0].headers[0].description.raw == desc

    res_response = res.responses[0].headers[0]
    res_type_resp = res.resource_type.responses[0].headers[0]
    assert res_response == res_type_resp

    res = resources[-10]

    assert res.display_name == "thingys"
    assert res.responses[0].code == 201
    assert res.responses[0].headers[0].name == "X-another-bogus-header"
    assert res.responses[0].headers[0].description.raw == "A bogus header"
    assert res.responses[0].body[0].mime_type == "application/json"
    assert res.responses[0].body[0].schema == "Thingy"


def test_resource_base_uri_params(resources):
    res = resources[2]

    assert res.display_name == "widget-gizmos"
    assert res.base_uri_params[0].name == "subdomain"

    # TODO: figure out what this issue is
    # desc = "subdomain for the baseUriType resource type"
    # assert res.base_uri_params[0].description.raw == desc
    # assert res.base_uri_params[0].default == "fooBar"

    res = resources[-12]
    assert len(res.base_uri_params) == 1
    assert res.display_name == "users-profile"
    assert res.base_uri_params[0].name == "subdomain"
    assert res.base_uri_params[0].default == "barFoo"

    desc = "a test base URI parameter for resource-level"
    assert res.base_uri_params[0].description.raw == desc


def test_resource_form_params(resources):
    res = resources[-3]

    assert res.display_name == "formParamResource"
    assert res.description.raw == "A example resource with form parameters"
    assert res.form_params[0].name == "foo"
    assert res.form_params[0].description.raw == "Post some foo"
    assert res.form_params[0].type == "string"
    assert res.form_params[0].required
    assert res.form_params[0].min_length == 10
    assert res.form_params[0].max_length == 100

    assert res.form_params[1].name == "bar"
    assert res.form_params[1].description.raw == "Post some bar"
    assert res.form_params[1].type == "string"
    assert res.form_params[1].required is False
    assert res.form_params[1].min_length == 15
    assert res.form_params[1].max_length == 150
    assert res.form_params[1].default == "aPostedBarExample"


def test_resource_security_scheme(resources):
    res = resources[17]
    assert res.method == "get"
    assert res.name == "/users/{user_id}/thingys"
    assert res.secured_by == [
        {"oauth_2_0": {"scopes": ["thingy-read-private"]}}
    ]
    assert res.security_schemes[0].name == "oauth_2_0"


def test_resource_inherit_parent(resources):
    res = resources[2]
    assert len(res.uri_params) == 2


def test_resource_response_no_desc():
    raml_file = os.path.join(EXAMPLES + "empty-mapping.raml")
    loaded_raml_file = load_file(raml_file)
    config = setup_config(EXAMPLES + "test-config.ini")
    config['validate'] = False
    api = pw.parse_raml(loaded_raml_file, config)

    res = api.resources[-1]
    response = res.responses[-1]

    assert response.code == 204
    assert response.description is None


@pytest.fixture(scope="session")
def inherited_resources():
    raml_file = os.path.join(EXAMPLES, "resource-type-inherited.raml")
    loaded_raml = load_file(raml_file)
    config = setup_config(EXAMPLES + "test-config.ini")
    config["validate"] = False
    return pw.parse_raml(loaded_raml, config)


def test_resource_inherits_type(inherited_resources):
    assert len(inherited_resources.resources) == 7
    res = inherited_resources.resources[0]
    assert res.type == "inheritgetmethod"
    assert res.method == "get"
    assert len(res.headers) == 1
    assert len(res.body) == 1
    assert len(res.responses) == 2
    assert len(res.query_params) == 1

    exp_desc = ("This description should be inherited when applied to "
                "resources")
    assert res.description.raw == exp_desc

    h = res.headers[0]
    assert h.name == "X-Inherited-Header"
    assert h.description.raw == "This header should be inherited"

    b = res.body[0]
    assert b.mime_type == "application/json"
    # lazy checking
    assert b.schema is not None
    assert b.example is not None

    q = res.query_params[0]
    assert q.name == "inherited_param"
    assert q.display_name == "inherited query parameter for a get method"
    assert q.type == "string"
    assert q.description.raw == "description for inherited query param"
    assert q.example == "fooBar"
    assert q.min_length == 1
    assert q.max_length == 50
    assert q.required is True


def test_res_type_inheritance(inherited_resources):
    res = inherited_resources.resource_types[0]
    assert res.name == "basetype"
    assert len(res.query_params) == 1
    assert len(res.form_params) == 1
    assert len(res.uri_params) == 1
    assert len(res.base_uri_params) == 1

    res = inherited_resources.resource_types[-1]
    assert res.name == "inheritbase"
    assert len(res.query_params) == 2
    assert len(res.form_params) == 2
    assert len(res.uri_params) == 2
    assert len(res.base_uri_params) == 2


def test_resource_inherits_type_optional_post(inherited_resources):
    assert len(inherited_resources.resources) == 7
    res = inherited_resources.resources[1]
    assert res.type == "inheritgetoptionalmethod"
    assert res.method == "post"
    assert res.headers is None
    assert res.query_params is None
    assert res.description.raw == "post some foobar"


def test_resource_inherits_type_optional_get(inherited_resources):
    # make sure that optional resource type methods are not inherited if not
    # explicitly included in resource (unless no methods defined)
    assert len(inherited_resources.resources) == 7
    res = inherited_resources.resources[2]
    assert res.type == "inheritgetoptionalmethod"
    assert res.method == "get"
    assert len(res.headers) == 2
    assert len(res.query_params) == 1

    desc = ("This description should be inherited when applied to resources "
            "with get methods")
    assert res.description.raw == desc


def test_resource_inherits_get(inherited_resources):
    assert len(inherited_resources.resources) == 7
    post_res = inherited_resources.resources[3]
    get_res = inherited_resources.resources[4]

    assert get_res.method == "get"
    assert len(get_res.headers) == 1
    assert len(get_res.body) == 1
    assert len(get_res.responses) == 2
    assert len(get_res.query_params) == 1

    h = get_res.headers[0]
    assert h.name == "X-Inherited-Header"
    assert h.description.raw == "This header should be inherited"

    b = get_res.body[0]
    assert b.mime_type == "application/json"
    # lazy checking
    assert b.schema is not None
    assert b.example is not None

    q = get_res.query_params[0]
    assert q.name == "inherited_param"
    assert q.display_name == "inherited query parameter for a get method"
    assert q.type == "string"
    assert q.description.raw == "description for inherited query param"
    assert q.example == "fooBar"
    assert q.min_length == 1
    assert q.max_length == 50
    assert q.required is True

    assert post_res.method == "post"
    assert post_res.description.raw == "post some more foobar"


def test_resource_inherited_no_overwrite(inherited_resources):
    # make sure that if resource inherits a resource type, and explicitly
    # defines properties that are defined in the resource type, the
    # properties in the resource take preference
    assert len(inherited_resources.resources) == 7
    res = inherited_resources.resources[5]

    assert res.method == "get"
    assert len(res.query_params) == 2

    desc = "This method-level description should be used"
    assert res.description.raw == desc

    # test query params
    first_param = res.query_params[0]
    second_param = res.query_params[1]

    assert second_param.name == "inherited"
    assert second_param.description.raw == "An inherited parameter"

    assert first_param.name == "overwritten"
    desc = "This query param description should be used"
    assert first_param.description.raw == desc

    # test form params
    first_param = res.form_params[0]

    desc = "This description should be inherited"
    assert first_param.description.raw == desc

    example = "This example for the overwritten form param should be used"
    assert first_param.example == example

    assert first_param.type == "string"
    assert first_param.min_length == 1
    assert first_param.max_length == 5

    # test headers
    first_header = res.headers[0]
    second_header = res.headers[1]

    assert first_header.name == "X-Inherited-Header"
    assert first_header.description.raw == "this header is inherited"

    assert second_header.name == "X-Overwritten-Header"
    desc = "This header description should be used"
    assert second_header.description.raw == desc
    assert second_header.required

    # test body
    first_body = res.body[0]
    second_body = res.body[1]

    assert first_body.mime_type == "application/json"

    schema = {
        "$schema": "http://json-schema.org/draft-03/schema",
        "type": "object",
        "properties": {
            "inherited": {
                "description": "this schema should be inherited"
            }
        }
    }
    example = {"inherited": "yes please!"}
    assert first_body.schema == schema
    assert first_body.example == example

    assert second_body.mime_type == "text/xml"

    schema = ("<xs:schema attributeFormDefault='unqualified' "
              "elementFormDefault='qualified' "
              "xmlns:xs='http://www.w3.org/2001/XMLSchema'>"
              "<xs:include schemaLocation='./thingy.xsd'/>"
              "<xs:element name='thingies'><xs:complexType>"
              "<xs:sequence minOccurs='0' maxOccurs='unbounded'>"
              "<xs:element type='thingyType' name='thingy'/>"
              "</xs:sequence></xs:complexType></xs:element></xs:schema>")
    schema = xmltodict.parse(schema)

    example = ("<thingies><thingy><name>Successfully overwrote body XML "
               "example</name></thingy></thingies>")
    example = xmltodict.parse(example)
    assert second_body.schema == schema
    assert second_body.example == example

    # test responses
    first_resp = res.responses[0]
    second_resp = res.responses[1]

    assert first_resp.code == 200

    desc = "overwriting the 200 response description"
    assert first_resp.description.raw == desc
    assert len(first_resp.headers) == 2

    first_header = first_resp.headers[0]
    second_header = first_resp.headers[1]

    assert first_header.name == "X-Inherited-Success"
    desc = "inherited success response header"
    assert first_header.description.raw == desc

    assert second_header.name == "X-Overwritten-Success"
    desc = "overwritten success response header"
    assert second_header.description.raw == desc

    assert second_resp.code == 201
    assert second_resp.body[0].mime_type == "application/json"
    example = {"description": "overwritten description of 201 body example"}
    assert second_resp.body[0].example == example


@pytest.fixture(scope="session")
def resource_protocol():
    raml_file = os.path.join(EXAMPLES, "protocols.raml")
    loaded_raml = load_file(raml_file)
    config = setup_config(EXAMPLES + "test-config.ini")
    config['validate'] = False
    return pw.parse_raml(loaded_raml, config)


def test_overwrite_protocol(resource_protocol):
    # if a resource explicitly defines a protocol, *that*
    # should be reflected in the absolute URI
    api = resource_protocol
    assert api.protocols == ["HTTPS"]
    assert api.base_uri == "https://api.spotify.com/v1"

    res = api.resources
    assert len(res) == 2

    first = res[0]
    second = res[1]

    assert first.display_name == "several-tracks"
    assert first.protocols == ["HTTP"]
    assert second.display_name == "track"
    assert second.protocols == ["HTTP"]


@pytest.fixture(scope="session")
def uri_param_resources():
    raml_file = os.path.join(EXAMPLES, "preserve-uri-order.raml")
    loaded_raml = load_file(raml_file)
    config = setup_config(EXAMPLES + "test-config.ini")
    config['validate'] = False
    return pw.parse_raml(loaded_raml, config)


def test_uri_params_order(uri_param_resources):
    res = uri_param_resources.resources[1]
    expected_uri = ["lang", "user_id", "playlist_id"]
    expected_base = ["subHostName", "bucketName"]

    uri = [u.name for u in res.uri_params]
    base = [b.name for b in res.base_uri_params]

    assert uri == expected_uri
    assert base == expected_base


@pytest.fixture(scope="session")
def undef_uri_params_resources():
    raml_file = os.path.join(EXAMPLES, "undefined-uri-params.raml")
    loaded_raml = load_file(raml_file)
    config = setup_config(EXAMPLES + "test-config.ini")
    config['validate'] = False
    return pw.parse_raml(loaded_raml, config)


def test_undefined_uri_params(undef_uri_params_resources):
    res = undef_uri_params_resources.resources[1]

    assert len(res.uri_params) == 1
    assert res.uri_params[0].name == "id"


#####
# Test Includes parsing
#####
@pytest.fixture(scope="session")
def xml_includes():
    raml_file = os.path.join(EXAMPLES + "xsd_includes.raml")
    loaded_raml_file = load_file(raml_file)
    config = setup_config(EXAMPLES + "test-config.ini")
    return pw.parse_raml(loaded_raml_file, config)


def test_includes_xml(xml_includes):
    api = xml_includes
    assert api.title == "Sample API Demo - XSD Includes"
    assert api.version == "v1"
    assert api.schemas == [{
        "xml": {
            "root": {
                "false": "true",
                "name": "foo",
            }
        },
    }]


@pytest.fixture(scope="session")
def json_includes():
    raml_file = os.path.join(EXAMPLES + "json_includes.raml")
    loaded_raml_file = load_file(raml_file)
    config = setup_config(EXAMPLES + "test-config.ini")
    return pw.parse_raml(loaded_raml_file, config)


def test_includes_json(json_includes):
    api = json_includes
    assert api.title == "Sample API Demo - JSON Includes"
    assert api.version == "v1"
    assert api.schemas == [{
        "json": {
            "false": True,
            "name": "foo",
        },
    }]


@pytest.fixture(scope="session")
def md_includes():
    raml_file = os.path.join(EXAMPLES + "md_includes.raml")
    loaded_raml_file = load_file(raml_file)
    config = setup_config(EXAMPLES + "test-config.ini")
    return pw.parse_raml(loaded_raml_file, config)


def test_includes_md(md_includes):
    api = md_includes
    assert api.title == "Sample API Demo - Markdown Includes"
    assert api.version == "v1"
    markdown_raw = """## Foo

*Bacon ipsum dolor* amet pork belly _doner_ rump brisket. [Cupim jerky \
shoulder][0] ball tip, jowl bacon meatloaf shank kielbasa turducken corned \
beef beef turkey porchetta.

### Doner meatball pork belly andouille drumstick sirloin

Porchetta picanha tail sirloin kielbasa, pig meatball short ribs drumstick \
jowl. Brisket swine spare ribs picanha t-bone. Ball tip beef tenderloin jowl \
doner andouille cupim meatball. Porchetta hamburger filet mignon jerky flank, \
meatball salami beef cow venison tail ball tip pork belly.

[0]: https://baconipsum.com/?paras=1&type=all-meat&start-with-lorem=1
"""
    assert api.documentation[0].content.raw == markdown_raw

    markdown_html = """<h2>Foo</h2>

<p><em>Bacon ipsum dolor</em> amet pork belly <em>doner</em> rump brisket. \
<a href="https://baconipsum.com/?paras=1&amp;type=all-meat&amp;start-with-\
lorem=1">Cupim jerky shoulder</a> ball tip, jowl bacon meatloaf shank \
kielbasa turducken corned beef beef turkey porchetta.</p>

<h3>Doner meatball pork belly andouille drumstick sirloin</h3>

<p>Porchetta picanha tail sirloin kielbasa, pig meatball short ribs drumstick \
jowl. Brisket swine spare ribs picanha t-bone. Ball tip beef tenderloin jowl \
doner andouille cupim meatball. Porchetta hamburger filet mignon jerky flank, \
meatball salami beef cow venison tail ball tip pork belly.</p>
"""
    assert api.documentation[0].content.html == markdown_html



# -*- coding: utf-8 -*-
# Copyright (c) 2015 Spotify AB
from __future__ import absolute_import, division, print_function

import os

PAR_DIR = os.path.abspath(os.path.dirname(__file__))
EXAMPLES = os.path.join(PAR_DIR + '/data/examples/')
VALIDATE = os.path.join(PAR_DIR + '/data/validate/')
FIXTURES = os.path.join(PAR_DIR + '/data/fixtures/')
UPDATE = os.path.join(PAR_DIR + '/data/update/')
JSONREF = os.path.join(PAR_DIR + '/data/jsonref/')



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2015 Spotify AB

from __future__ import absolute_import, division, print_function

import click

from .tree import tree as ttree
from .errors import InvalidRAMLError
from .utils import update_mime_types as umt
from ._helpers import load_file

from ramlfications import validate as vvalidate


@click.group()
def main():
    """Yet Another RAML Parser"""
    # Needed to collect the validate & tree commands


@main.command(help="Validate a RAML file.")
@click.argument("ramlfile", type=click.Path(exists=True))
@click.option("--config", "-c", type=click.Path(exists=True),
              help="Additionally supported items beyond RAML spec.")
def validate(ramlfile, config):
    """Validate a given RAML file."""
    try:
        vvalidate(ramlfile, config)
        click.secho("Success! Valid RAML file: {0}".format(ramlfile),
                    fg="green")

    except InvalidRAMLError as e:
        msg = "Error validating file {0}: \n{1}".format(ramlfile, e)
        click.secho(msg, fg="red", err=True)
        raise SystemExit(1)


@main.command(help="Visualize the RAML file as a tree.")
@click.argument('ramlfile', type=click.Path(exists=True))
@click.option("-C", "--color", type=click.Choice(['dark', 'light']),
              default=None,
              help=("Color theme 'light' for dark-screened backgrounds"))
@click.option("-o", "--output", type=click.File('w'),
              help=("Save tree output to file"))
@click.option("-v", "--verbose", default=0, count=True,
              help="Include methods for each endpoint")
@click.option("-V", "--validate", default=False, is_flag=True,
              help="Validate RAML file")
@click.option("-c", "--config", type=click.Path(exists=True),
              help="Additionally supported items beyond RAML spec.")
def tree(ramlfile, color, output, verbose, validate, config):
    """Pretty-print a tree of the RAML-defined API."""
    try:
        load_obj = load_file(ramlfile)
        ttree(load_obj, color, output, verbose, validate, config)
    except InvalidRAMLError as e:
        msg = '"{0}" is not a valid RAML file: {1}'.format(
            click.format_filename(ramlfile), e)
        click.secho(msg, fg="red", err=True)
        raise SystemExit(1)


@main.command(help="Update RAMLfications' supported MIME types from IANA.")
def update():
    umt()


if __name__ == "__main__":
    main()



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2015 Spotify AB
from __future__ import absolute_import, division, print_function


from termcolor import colored
# NOTE: I know that these should be caps being constants but I'm
# not liking that for some reason for fixtures. Deal.

tree_no_color = ("""\
==================================
Spotify Web API Demo - Simple Tree
==================================
Base URI: https://api.spotify.com/v1
|- /tracks
|  - /tracks/{id}
|- /users/{user_id}/playlists
|  - /users/{user_id}/playlists/{playlist_id}
""")

# NOTE: Since I am having the pipe ("|") consistent throughout in terms
# of color, as well as setting its ANSI color separately from the rest
# of the string - even if they are the same color - tests need to have
# call `colored()` around the pipe separately.

tree_light = (
    colored("==================================", "white") + "\n" +
    colored("Spotify Web API Demo - Simple Tree", "yellow") + "\n" +
    colored("==================================", "white") + "\n" +
    colored("Base URI: https://api.spotify.com/v1", "yellow") + "\n" +
    colored("|", "green", attrs=["bold"]) +
    colored("- /tracks", "green", attrs=["bold"]) + "\n" +
    colored("|", "green", attrs=["bold"]) +
    colored("  - /tracks/{id}", "green", attrs=["bold"]) + "\n" +
    colored("|", "green", attrs=["bold"]) +
    colored("- /users/{user_id}/playlists", "green", attrs=["bold"]) + "\n" +
    colored("|", "green", attrs=["bold"]) +
    colored("  - /users/{user_id}/playlists/{playlist_id}", "green",
            attrs=["bold"]) + "\n"
)

tree_dark = (
    colored("==================================", "grey") + "\n" +
    colored("Spotify Web API Demo - Simple Tree", "grey",
            attrs=["bold"]) + "\n" +
    colored("==================================", "grey") + "\n" +
    colored("Base URI: https://api.spotify.com/v1",
            "grey", attrs=["bold"]) + "\n" +
    colored("|", "grey", attrs=["bold"]) +
    colored("- /tracks", "grey", attrs=["bold"]) + "\n" +
    colored("|", "grey", attrs=["bold"]) +
    colored("  - /tracks/{id}", "grey", attrs=["bold"]) + "\n" +
    colored("|", "grey", attrs=["bold"]) +
    colored("- /users/{user_id}/playlists", "grey", attrs=["bold"]) + "\n" +
    colored("|", "grey", attrs=["bold"]) +
    colored("  - /users/{user_id}/playlists/{playlist_id}", "grey",
            attrs=["bold"]) + "\n"
)

tree_light_v = (
    colored("==================================", "white") + "\n" +
    colored("Spotify Web API Demo - Simple Tree", "yellow") + "\n" +
    colored("==================================", "white") + "\n" +
    colored("Base URI: https://api.spotify.com/v1", "yellow") + "\n" +
    colored("|", "green", attrs=["bold"]) +
    colored("- /tracks", "green", attrs=["bold"]) + "\n" +
    colored("|", "green", attrs=["bold"]) +
    colored("  ⌙ GET", "blue", attrs=["bold"]) + "\n" +
    colored("|", "green", attrs=["bold"]) +
    colored("  - /tracks/{id}", "green", attrs=["bold"]) + "\n" +
    colored("|", "green", attrs=["bold"]) +
    colored("    ⌙ GET", "blue", attrs=["bold"]) + "\n" +
    colored("|", "green", attrs=["bold"]) +
    colored("- /users/{user_id}/playlists", "green",
            attrs=["bold"]) + "\n" +
    colored("|", "green", attrs=["bold"]) +
    colored("  ⌙ GET", "blue", attrs=["bold"]) + "\n" +
    colored("|", "green", attrs=["bold"]) +
    colored("  - /users/{user_id}/playlists/{playlist_id}", "green",
            attrs=["bold"]) + "\n" +
    colored("|", "green", attrs=["bold"]) +
    colored("    ⌙ PUT", "blue", attrs=["bold"]) + "\n"
)

tree_light_vv = (
    colored("==================================", "white") + "\n" +
    colored("Spotify Web API Demo - Simple Tree", "yellow") + "\n" +
    colored("==================================", "white") + "\n" +
    colored("Base URI: https://api.spotify.com/v1", "yellow") + "\n" +
    colored("|", "green", attrs=["bold"]) +
    colored("- /tracks", "green", attrs=["bold"]) + "\n" +
    colored("|", "green", attrs=["bold"]) +
    colored("  ⌙ GET", "blue", attrs=["bold"]) + "\n" +
    colored("|", "green", attrs=["bold"]) +
    colored("     Query Params", "cyan", attrs=["bold"]) + "\n" +
    colored("|", "green", attrs=["bold"]) +
    colored("      ⌙ ids", "cyan", attrs=["bold"]) + "\n" +
    colored("|", "green", attrs=["bold"]) +
    colored("  - /tracks/{id}", "green", attrs=["bold"]) + "\n" +
    colored("|", "green", attrs=["bold"]) +
    colored("    ⌙ GET", "blue", attrs=["bold"]) + "\n" +
    colored("|", "green", attrs=["bold"]) +
    colored("       URI Params", "cyan", attrs=["bold"]) + "\n" +
    colored("|", "green", attrs=["bold"]) +
    colored("        ⌙ id", "cyan", attrs=["bold"]) + "\n" +
    colored("|", "green", attrs=["bold"]) +
    colored("- /users/{user_id}/playlists", "green", attrs=["bold"]) + "\n" +
    colored("|", "green", attrs=["bold"]) +
    colored("  ⌙ GET", "blue", attrs=["bold"]) + "\n" +
    colored("|", "green", attrs=["bold"]) +
    colored("     URI Params", "cyan", attrs=["bold"]) + "\n" +
    colored("|", "green", attrs=["bold"]) +
    colored("      ⌙ user_id", "cyan", attrs=["bold"]) + "\n" +
    colored("|", "green", attrs=["bold"]) +
    colored("  - /users/{user_id}/playlists/{playlist_id}", "green",
            attrs=["bold"]) + "\n" +
    colored("|", "green", attrs=["bold"]) +
    colored("    ⌙ PUT", "blue", attrs=["bold"]) + "\n" +
    colored("|", "green", attrs=["bold"]) +
    colored("       URI Params", "cyan", attrs=["bold"]) + "\n" +
    colored("|", "green", attrs=["bold"]) +
    colored("        ⌙ user_id", "cyan", attrs=["bold"]) + "\n" +
    colored("|", "green", attrs=["bold"]) +
    colored("       Form Params", "cyan", attrs=["bold"]) + "\n" +
    colored("|", "green", attrs=["bold"]) +
    colored("        ⌙ name", "cyan", attrs=["bold"]) + "\n"
)

tree_light_vvv = (
    colored("==================================", "white") + "\n" +
    colored("Spotify Web API Demo - Simple Tree", "yellow") + "\n" +
    colored("==================================", "white") + "\n" +
    colored("Base URI: https://api.spotify.com/v1", "yellow") + "\n" +
    colored("|", "green", attrs=["bold"]) +
    colored("- /tracks", "green", attrs=["bold"]) + "\n" +
    colored("|", "green", attrs=["bold"]) +
    colored("  ⌙ GET", "blue", attrs=["bold"]) + "\n" +
    colored("|", "green", attrs=["bold"]) +
    colored("     Query Params", "cyan", attrs=["bold"]) + "\n" +
    colored("|", "green", attrs=["bold"]) +
    colored("      ⌙ ids: Spotify Track IDs", "cyan", attrs=["bold"]) + "\n" +
    colored("|", "green", attrs=["bold"]) +
    colored("  - /tracks/{id}", "green", attrs=["bold"]) + "\n" +
    colored("|", "green", attrs=["bold"]) +
    colored("    ⌙ GET", "blue", attrs=["bold"]) + "\n" +
    colored("|", "green", attrs=["bold"]) +
    colored("       URI Params", "cyan", attrs=["bold"]) + "\n" +
    colored("|", "green", attrs=["bold"]) +
    colored("        ⌙ id: Spotify Track ID", "cyan", attrs=["bold"]) + "\n" +
    colored("|", "green", attrs=["bold"]) +
    colored("- /users/{user_id}/playlists", "green", attrs=["bold"]) + "\n" +
    colored("|", "green", attrs=["bold"]) +
    colored("  ⌙ GET", "blue", attrs=["bold"]) + "\n" +
    colored("|", "green", attrs=["bold"]) +
    colored("     URI Params", "cyan", attrs=["bold"]) + "\n" +
    colored("|", "green", attrs=["bold"]) +
    colored("      ⌙ user_id: User ID", "cyan", attrs=["bold"]) + "\n" +
    colored("|", "green", attrs=["bold"]) +
    colored("  - /users/{user_id}/playlists/{playlist_id}", "green",
            attrs=["bold"]) + "\n" +
    colored("|", "green", attrs=["bold"]) +
    colored("    ⌙ PUT", "blue", attrs=["bold"]) + "\n" +
    colored("|", "green", attrs=["bold"]) +
    colored("       URI Params", "cyan", attrs=["bold"]) + "\n" +
    colored("|", "green", attrs=["bold"]) +
    colored("        ⌙ user_id: User ID", "cyan", attrs=["bold"]) + "\n" +
    colored("|", "green", attrs=["bold"]) +
    colored("       Form Params", "cyan", attrs=["bold"]) + "\n" +
    colored("|", "green", attrs=["bold"]) +
    colored("        ⌙ name: Playlist Name", "cyan", attrs=["bold"]) + "\n"
)



from .errors import BaseRAMLError


def collecterrors(func):
    def func_wrapper(inst, attr, value):
        try:
            func(inst, attr, value)
        except BaseRAMLError as e:
            inst.errors.append(e)

    return func_wrapper



# -*- coding: utf-8 -*-
# Copyright (c) 2015 Spotify AB

from __future__ import absolute_import, division, print_function

try:
    from collections import OrderedDict
except ImportError:  # pragma: no cover
    from ordereddict import OrderedDict  # pragma: no cover

import sys

from six import iteritems, itervalues
from termcolor import colored

from .config import setup_config
from .parser import parse_raml

COLOR_MAP = {
    "light": (('white', None),
              ('yellow', None),
              ('green', 'bold'),
              ('blue', 'bold'),
              ('cyan', 'bold')),
    "dark": (('grey', None),
             ('grey', 'bold'),
             ('grey', 'bold'),
             ('magenta', None),
             ('blue', None))
}


def _get_node_level(node, count):
    if node.parent:
        count += 1
        return _get_node_level(node.parent, count)
    else:
        return count


def _get_tree(api):
    resources = OrderedDict()
    for r in api.resources:
        if r.method is not None:
            key = r.method.upper() + " " + r.path
        else:
            key = r.path
        resources[key] = r

    return resources


def _create_space(v):
    space_count = _get_node_level(v, count=0)
    space = "  " * space_count
    return space


def _set_ansi(string, screen_color, line_color):
    if screen_color:
        color, attr = COLOR_MAP[screen_color][line_color]
        if attr:
            return colored(string, color, attrs=[attr])
        return colored(string, color)
    return string


def _print_line(sp, msg, color, line_color):
    pipe = _set_ansi("|", color, 2)
    output = pipe + _set_ansi(sp + msg, color, line_color) + "\n"
    sys.stdout.write(output)


def _return_param_type(node):
    params = OrderedDict()
    if node.query_params:
        params["Query Params"] = node.query_params
    if node.uri_params:
        params["URI Params"] = node.uri_params
    if node.form_params:
        params["Form Params"] = node.form_params
    return params


def _params(space, node, color, desc):
    params = _return_param_type(node)
    for k, v in list(iteritems(params)):
        _print_line(space + '     ', k, color, 4)
        for p in v:
            if desc:
                desc = ": " + p.display_name
            else:
                desc = ''
            _print_line(space + '      ⌙ ', p.name + desc, color, 4)


def _print_verbosity(resources, color, verbosity):
    for r in list(itervalues(resources)):
        space = _create_space(r)
        _print_line(space, "- " + r.path, color, 2)
        if verbosity > 0:
            if r.method:
                _print_line(space, "  ⌙ " + r.method.upper(), color, 3)
            else:
                continue
            if verbosity > 1:
                desc = verbosity == 3
                _params(space, r, color, desc)


def _print_metadata(api, color):
    if not api.title:
        api.title = "MISSING TITLE"
    head = _set_ansi("=" * len(api.title), color, 0) + "\n"
    head += _set_ansi(api.title, color, 1) + "\n"
    head += _set_ansi("=" * len(api.title), color, 0) + "\n"
    head += _set_ansi("Base URI: " + api.base_uri, color, 1) + "\n"

    sys.stdout.write(head)


def _print_tree(api, ordered_resources, print_color, verbosity):
    _print_metadata(api, print_color)
    _print_verbosity(ordered_resources, print_color, verbosity)


def tree(load_obj, color, output, verbosity, validate, config):  # NOCOV
    """
    Create a tree visualization of given RAML file.

    :param dict load_obj: Loaded RAML File
    :param str color: ``light``, ``dark`` or ``None`` (default) for the color
        output
    :param str output: Path to output file, if given
    :param str verbosity: Level of verbosity to print out
    :return: ASCII Tree representation of API
    :rtype: stdout to screen or given file name
    :raises InvalidRootNodeError: API metadata is invalid according to RAML \
        `specification <http://raml.org/spec.html>`_.
    :raises InvalidResourceNodeError: API resource endpoint is invalid \
        according to RAML `specification <http://raml.org/spec.html>`_.
    :raises InvalidParameterError: Named parameter is invalid \
        according to RAML `specification <http://raml.org/spec.html>`_.
    """
    config = setup_config(config)
    api = parse_raml(load_obj, config)
    resources = _get_tree(api)

    if output:
        sys.stdout = output  # file is opened via @click from __main__.py
        _print_tree(api, resources, color, verbosity)
        output.close()
    else:
        _print_tree(api, resources, color, verbosity)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2015 Spotify AB
import io
import os
import re
import sys
from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand  # NOQA


NAME = "ramlfications"
META_PATH = os.path.join("ramlfications", "__init__.py")


HERE = os.path.abspath(os.path.dirname(__file__))


def read(*filenames, **kwargs):
    encoding = kwargs.get('encoding', 'utf-8')
    sep = kwargs.get('sep', '\n')
    buf = []
    for fl in filenames:
        with io.open(fl, encoding=encoding) as f:
            buf.append(f.read())
    return sep.join(buf)


META_FILE = read(META_PATH)


def find_meta(meta):
    """
    Extract __*meta*__ from META_FILE.
    """
    meta_match = re.search(
        r"^__{meta}__ = ['\"]([^'\"]*)['\"]".format(meta=meta),
        META_FILE, re.M
    )
    if meta_match:
        return meta_match.group(1)
    raise RuntimeError("Unable to find __{meta}__ string.".format(meta=meta))


def install_requires():
    install_requires = [
        "attrs", "click", "jsonref", "markdown2", "pyyaml", "six",
        "termcolor", "xmltodict"
    ]
    if sys.version_info[:2] == (2, 6):
        install_requires.append("ordereddict")
    return install_requires


long_description = read('README.rst', 'docs/changelog.rst')


class PyTest(TestCommand):
    user_options = [("pytest-args=", "a", "Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = ''

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest
        errno = pytest.main(self.pytest_args + ' tests')
        sys.exit(errno)

setup(
    name=NAME,
    version=find_meta("version"),
    description=find_meta("description"),
    long_description=long_description,
    url=find_meta("uri"),
    license=find_meta("license"),
    author=find_meta("author"),
    author_email=find_meta("email"),
    keywords=["raml", "rest"],
    packages=find_packages(exclude=["tests*"]),
    entry_points={
        'console_scripts': [
            'ramlfications = ramlfications.__main__:main'
        ]
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    package_data={"": ["data/supported_mime_types.json"]},
    install_requires=install_requires(),
    extras_require={
        "all": ["requests[security]"]
    },
    tests_require=[
        "pytest", "mock"
    ],
    cmdclass={
        "test": PyTest,
    }
)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2014 Spotify AB
from __future__ import absolute_import, division, print_function

import os

import pytest

from ramlfications import parse
from ramlfications import parser as pw
from ramlfications.config import setup_config
from ramlfications.raml import RootNode, ResourceTypeNode, TraitNode
from ramlfications._helpers import load_file

from tests.base import EXAMPLES


@pytest.fixture(scope="session")
def github_raml():
    raml_file = os.path.join(EXAMPLES, "github.raml")
    return load_file(raml_file)


def test_parse_raml(github_raml):
    config_file = os.path.join(EXAMPLES, "github-config.ini")
    config = setup_config(config_file)
    root = pw.parse_raml(github_raml, config)
    assert isinstance(root, RootNode)


@pytest.fixture(scope="session")
def root():
    raml_file = os.path.join(EXAMPLES, "github.raml")
    loaded_raml_file = load_file(raml_file)
    config_file = os.path.join(EXAMPLES, "github-config.ini")
    config = setup_config(config_file)
    return pw.create_root(loaded_raml_file, config)


def test_base_uri(root):
    expected = "https://api.github.com/"
    assert expected == root.base_uri


def test_protocols(root):
    expected = ["HTTPS"]
    assert expected == root.protocols


def test_docs(root):
    assert root.documentation is None


def test_base_uri_params(root):
    assert root.base_uri_params is None


def test_uri_params(root):
    assert root.uri_params is None


def test_title(root):
    assert root.title == "GitHub API"


def test_version(root):
    assert root.version == "v3"


def test_schemas(root):
    assert root.schemas is None


def test_media_type(root):
    assert root.media_type == "application/json"


#####
# Security Schemes
#####

@pytest.fixture(scope="session")
def sec_schemes():
    raml_file = os.path.join(EXAMPLES, "github.raml")
    config = os.path.join(EXAMPLES, "github-config.ini")
    api = parse(raml_file, config)
    return api.security_schemes


def test_create_security_schemes(sec_schemes):
    assert len(sec_schemes) == 2
    assert sec_schemes[0].name == "basic"
    assert sec_schemes[0].type == "Basic Authentication"

    assert sec_schemes[1].name == "oauth_2_0"
    assert sec_schemes[1].type == "OAuth 2.0"

    desc = ("OAuth2 is a protocol that lets external apps request "
            "authorization to private\ndetails in a user's GitHub "
            "account without getting their password. This is\npreferred "
            "over Basic Authentication because tokens can be limited to "
            "specific\ntypes of data, and can be revoked by users at "
            "any time.\n")
    assert sec_schemes[1].description.raw == desc

    assert len(sec_schemes[1].headers) == 1
    assert len(sec_schemes[1].query_params) == 1
    assert len(sec_schemes[1].responses) == 1

    assert sec_schemes[1].headers[0].name == "Authorization"
    assert sec_schemes[1].headers[0].type == "string"

    desc = ("Used to send a valid OAuth 2 access token. Do not use together "
            "with\nthe \"access_token\" query string parameter.\n")
    assert sec_schemes[1].headers[0].description.raw == desc

    assert sec_schemes[1].query_params[0].name == "access_token"
    assert sec_schemes[1].query_params[0].type == "string"

    desc = ("Used to send a valid OAuth 2 access token. Do not use "
            "together with\nthe \"Authorization\" header\n")
    assert sec_schemes[1].query_params[0].description.raw == desc

    assert sec_schemes[1].responses[0].code == 404
    assert sec_schemes[1].responses[0].description.raw == "Unauthorized"

    settings = {
        "authorizationUri": "https://github.com/login/oauth/authorize",
        "accessTokenUri": "https://github.com/login/oauth/access_token",
        "authorizationGrants": ["code"],
        "scopes": [
            "user",
            "user:email",
            "user:follow",
            "public_repo",
            "repo",
            "repo:status",
            "delete_repo",
            "notifications",
            "gist"
        ]
    }
    assert sec_schemes[1].settings == settings


#####
# Traits
#####

@pytest.fixture(scope="session")
def traits():
    raml_file = os.path.join(EXAMPLES, "github.raml")
    config = os.path.join(EXAMPLES, "github-config.ini")
    api = parse(raml_file, config)
    return api.traits


def test_create_traits(traits):
    for trait in traits:
        assert isinstance(trait, TraitNode)

    assert len(traits) == 2


def test_trait_historical(traits):
    trait = traits[0]

    assert trait.form_params is None
    assert trait.uri_params is None
    assert trait.headers is None
    assert trait.body is None
    assert trait.responses is None
    assert len(trait.query_params) == 1

    assert trait.name == "historical"
    assert trait.query_params[0].name == "since"
    assert trait.query_params[0].type == "string"

    desc = ("Timestamp in ISO 8601 format YYYY-MM-DDTHH:MM:SSZ.\n"
            "Only gists updated at or after this time are returned.\n")
    assert trait.query_params[0].description.raw == desc


def test_trait_filterable(traits):
    trait = traits[1]

    assert trait.name == "filterable"
    assert trait.form_params is None
    assert trait.uri_params is None
    assert trait.headers is None
    assert trait.body is None
    assert trait.responses is None

    assert len(trait.query_params) == 6

    filter_ = trait.query_params[0]
    assert filter_.name == "filter"
    assert filter_.enum == [
        "assigned", "created", "mentioned", "subscribed", "all"
    ]
    assert filter_.default == "all"
    assert filter_.required
    assert filter_.type == "string"
    desc = ("Issues assigned to you / created by you / mentioning you / "
            "you're\nsubscribed to updates for / All issues the authenticated "
            "user can see\n")
    assert filter_.description.raw == desc

    state = trait.query_params[1]
    assert state.name == "state"
    assert state.enum == ["open", "closed"]
    assert state.default == "open"
    assert state.required
    assert state.type == "string"
    assert state.description is None
    assert not hasattr(state.description, "raw")
    assert not hasattr(state.description, "html")

    label = trait.query_params[2]
    assert label.name == "labels"
    assert label.type == "string"
    assert label.required
    desc = ("String list of comma separated Label names. Example - "
            "bug,ui,@high.")
    assert label.description.raw == desc

    sort = trait.query_params[3]
    assert sort.name == "sort"
    assert sort.enum == ["created", "updated", "comments"]
    assert sort.default == "created"
    assert sort.required
    assert sort.type == "string"
    assert sort.description is None
    assert not hasattr(sort.description, "raw")
    assert not hasattr(sort.description, "html")

    direction = trait.query_params[4]
    assert direction.name == "direction"
    assert direction.enum == ["asc", "desc"]
    assert direction.default == "desc"
    assert direction.required
    assert direction.description is None
    assert not hasattr(direction.description, "raw")
    assert not hasattr(direction.description, "html")

    since = trait.query_params[5]
    assert since.name == "since"
    assert since.type == "string"
    desc = ("Optional string of a timestamp in ISO 8601 format: "
            "YYYY-MM-DDTHH:MM:SSZ.\nOnly issues updated at or after this "
            "time are returned.\n")
    assert since.description.raw == desc


#####
# Resource Types
#####

@pytest.fixture(scope="session")
def resource_types():
    raml_file = os.path.join(EXAMPLES, "github.raml")
    config = os.path.join(EXAMPLES, "github-config.ini")
    api = parse(raml_file, config)
    return api.resource_types


def test_create_resource_types(resource_types):
    for r in resource_types:
        assert isinstance(r, ResourceTypeNode)

    assert len(resource_types) == 12


def test_resource_type_get_base(resource_types):
    res = resource_types[0]

    assert res.method == "get"
    assert len(res.headers) == 6
    assert len(res.responses) == 1
    assert res.body is None
    assert res.query_params is None
    assert res.form_params is None
    assert res.uri_params is None
    assert res.optional

    header = res.headers[0]
    assert header.name == "X-GitHub-Media-Type"
    assert header.type == "string"
    desc = "You can check the current version of media type in responses.\n"
    assert header.description.raw == desc

    header = res.headers[1]
    assert header.name == "Accept"
    assert header.type == "string"
    assert header.description.raw == "Is used to set specified media type."

    header = res.headers[2]
    assert header.name == "X-RateLimit-Limit"
    assert header.type == "integer"

    header = res.headers[3]
    assert header.name == "X-RateLimit-Remaining"
    assert header.type == "integer"

    header = res.headers[4]
    assert header.name == "X-RateLimit-Reset"
    assert header.type == "integer"

    header = res.headers[5]
    assert header.name == "X-GitHub-Request-Id"
    assert header.type == "integer"

    assert res.responses[0].code == 403
    desc = ("API rate limit exceeded. "
            "See http://developer.github.com/v3/#rate-limiting\nfor "
            "details.\n")
    assert res.responses[0].description.raw == desc


def test_resource_type_post_base(resource_types):
    res = resource_types[1]

    assert res.method == "post"
    assert len(res.headers) == 6
    assert len(res.responses) == 1
    assert res.body is None
    assert res.query_params is None
    assert res.form_params is None
    assert res.uri_params is None
    assert res.optional

    header = res.headers[0]
    assert header.name == "X-GitHub-Media-Type"
    assert header.type == "string"
    desc = "You can check the current version of media type in responses.\n"
    assert header.description.raw == desc

    header = res.headers[1]
    assert header.name == "Accept"
    assert header.type == "string"
    assert header.description.raw == "Is used to set specified media type."

    header = res.headers[2]
    assert header.name == "X-RateLimit-Limit"
    assert header.type == "integer"

    header = res.headers[3]
    assert header.name == "X-RateLimit-Remaining"
    assert header.type == "integer"

    header = res.headers[4]
    assert header.name == "X-RateLimit-Reset"
    assert header.type == "integer"

    header = res.headers[5]
    assert header.name == "X-GitHub-Request-Id"
    assert header.type == "integer"

    assert res.responses[0].code == 403
    desc = ("API rate limit exceeded. "
            "See http://developer.github.com/v3/#rate-limiting\nfor "
            "details.\n")
    assert res.responses[0].description.raw == desc


def test_resource_type_patch_base(resource_types):
    res = resource_types[2]

    assert res.method == "patch"
    assert len(res.headers) == 6
    assert len(res.responses) == 1
    assert res.body is None
    assert res.query_params is None
    assert res.form_params is None
    assert res.uri_params is None
    assert res.optional

    header = res.headers[0]
    assert header.name == "X-GitHub-Media-Type"
    assert header.type == "string"
    desc = "You can check the current version of media type in responses.\n"
    assert header.description.raw == desc

    header = res.headers[1]
    assert header.name == "Accept"
    assert header.type == "string"
    assert header.description.raw == "Is used to set specified media type."

    header = res.headers[2]
    assert header.name == "X-RateLimit-Limit"
    assert header.type == "integer"

    header = res.headers[3]
    assert header.name == "X-RateLimit-Remaining"
    assert header.type == "integer"

    header = res.headers[4]
    assert header.name == "X-RateLimit-Reset"
    assert header.type == "integer"

    header = res.headers[5]
    assert header.name == "X-GitHub-Request-Id"
    assert header.type == "integer"

    assert res.responses[0].code == 403
    desc = ("API rate limit exceeded. "
            "See http://developer.github.com/v3/#rate-limiting\nfor "
            "details.\n")
    assert res.responses[0].description.raw == desc


def test_resource_type_put_base(resource_types):
    res = resource_types[3]

    assert res.method == "put"
    assert len(res.headers) == 6
    assert len(res.responses) == 1
    assert res.body is None
    assert res.query_params is None
    assert res.form_params is None
    assert res.uri_params is None
    assert res.optional

    header = res.headers[0]
    assert header.name == "X-GitHub-Media-Type"
    assert header.type == "string"
    desc = "You can check the current version of media type in responses.\n"
    assert header.description.raw == desc

    header = res.headers[1]
    assert header.name == "Accept"
    assert header.type == "string"
    assert header.description.raw == "Is used to set specified media type."

    header = res.headers[2]
    assert header.name == "X-RateLimit-Limit"
    assert header.type == "integer"

    header = res.headers[3]
    assert header.name == "X-RateLimit-Remaining"
    assert header.type == "integer"

    header = res.headers[4]
    assert header.name == "X-RateLimit-Reset"
    assert header.type == "integer"

    header = res.headers[5]
    assert header.name == "X-GitHub-Request-Id"
    assert header.type == "integer"

    assert res.responses[0].code == 403
    desc = ("API rate limit exceeded. "
            "See http://developer.github.com/v3/#rate-limiting\nfor "
            "details.\n")
    assert res.responses[0].description.raw == desc


def test_resource_type_delete_base(resource_types):
    res = resource_types[4]

    assert res.method == "delete"
    assert len(res.headers) == 6
    assert len(res.responses) == 1
    assert res.body is None
    assert res.query_params is None
    assert res.form_params is None
    assert res.uri_params is None
    assert res.optional

    header = res.headers[0]
    assert header.name == "X-GitHub-Media-Type"
    assert header.type == "string"
    desc = "You can check the current version of media type in responses.\n"
    assert header.description.raw == desc

    header = res.headers[1]
    assert header.name == "Accept"
    assert header.type == "string"
    assert header.description.raw == "Is used to set specified media type."

    header = res.headers[2]
    assert header.name == "X-RateLimit-Limit"
    assert header.type == "integer"

    header = res.headers[3]
    assert header.name == "X-RateLimit-Remaining"
    assert header.type == "integer"

    header = res.headers[4]
    assert header.name == "X-RateLimit-Reset"
    assert header.type == "integer"

    header = res.headers[5]
    assert header.name == "X-GitHub-Request-Id"
    assert header.type == "integer"

    assert res.responses[0].code == 403
    desc = ("API rate limit exceeded. "
            "See http://developer.github.com/v3/#rate-limiting\nfor "
            "details.\n")
    assert res.responses[0].description.raw == desc


def test_resource_type_get_item(resource_types):
    res = [r for r in resource_types if r.name == "item"]
    res = [r for r in res if r.method == "get"][0]

    assert res.name == "item"
    assert res.type == "base"
    assert res.method == "get"
    assert len(res.headers) == 6
    assert len(res.responses) == 1
    assert res.body is None
    assert res.query_params is None
    assert res.form_params is None
    assert res.uri_params is None
    assert res.optional

    def _sort_headers(item):
        return item.name

    headers = sorted(res.headers, key=_sort_headers)

    header = headers[0]
    assert header.name == "Accept"
    assert header.type == "string"
    assert header.description.raw == "Is used to set specified media type."

    header = headers[1]
    assert header.name == "X-GitHub-Media-Type"
    assert header.type == "string"
    desc = "You can check the current version of media type in responses.\n"
    assert header.description.raw == desc

    header = headers[2]
    assert header.name == "X-GitHub-Request-Id"
    assert header.type == "integer"

    header = headers[3]
    assert header.name == "X-RateLimit-Limit"
    assert header.type == "integer"

    header = headers[4]
    assert header.name == "X-RateLimit-Remaining"
    assert header.type == "integer"

    header = headers[5]
    assert header.name == "X-RateLimit-Reset"
    assert header.type == "integer"

    assert res.responses[0].code == 403
    desc = ("API rate limit exceeded. "
            "See http://developer.github.com/v3/#rate-limiting\nfor "
            "details.\n")
    assert res.responses[0].description.raw == desc



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2014 Spotify AB
from __future__ import absolute_import, division, print_function

import os

import pytest

from ramlfications import parser as pw
from ramlfications.config import setup_config
from ramlfications.raml import (
    RootNode, ResourceTypeNode, TraitNode, ResourceNode
)
from ramlfications._helpers import load_file

from tests.base import EXAMPLES


@pytest.fixture(scope="session")
def raml():
    raml_file = os.path.join(EXAMPLES, "twitter.raml")
    return load_file(raml_file)


def test_parse_raml(raml):
    config_file = os.path.join(EXAMPLES, "twitter-config.ini")
    config = setup_config(config_file)
    root = pw.parse_raml(raml, config)
    assert isinstance(root, RootNode)


@pytest.fixture(scope="session")
def root(raml):
    config_file = os.path.join(EXAMPLES, "twitter-config.ini")
    config = setup_config(config_file)
    return pw.create_root(raml, config)


@pytest.fixture(scope="session")
def api(raml):
    config_file = os.path.join(EXAMPLES, "twitter-config.ini")
    config = setup_config(config_file)
    return pw.parse_raml(raml, config)


def test_base_uri(root):
    expected = "https://api.twitter.com/1.1"
    assert expected == root.base_uri


def test_protocols(root):
    expected = ["HTTPS"]
    assert expected == root.protocols


def test_docs(root):
    assert root.documentation is None


def test_base_uri_params(root):
    assert root.base_uri_params is None


def test_uri_params(root):
    assert root.uri_params is None


def test_title(root):
    assert root.title == "Twitter API"


def test_version(root):
    assert root.version == "1.1"


def test_schemas(root):
    assert root.schemas is None


def test_media_type(root):
    assert root.media_type == "application/json"


def test_secured_by(root):
    assert root.secured_by == ["oauth_1_0"]


#####
# Security Schemes
#####

@pytest.fixture
def sec_schemes(api):
    return api.security_schemes


def test_create_security_schemes(sec_schemes):
    assert len(sec_schemes) == 1
    assert sec_schemes[0].name == "oauth_1_0"
    assert sec_schemes[0].type == "OAuth 1.0"
    desc = ("Twitter offers applications the ability to issue authenticated "
            "requests on\nbehalf of the application itself (as opposed to "
            "on behalf of a specific user).\n")
    assert sec_schemes[0].description.raw == desc
    settings = {
        "requestTokenUri": "https://api.twitter.com/oauth/request_token",
        "authorizationUri": "https://api.twitter.com/oauth/authorize",
        "tokenCredentialsUri": "https://api.twitter.com/oauth/access_token"
    }
    assert sec_schemes[0].settings == settings


@pytest.fixture
def resource_types(api):
    return api.resource_types


def test_resource_types(resource_types):
    for r in resource_types:
        assert isinstance(r, ResourceTypeNode)
    assert len(resource_types) == 2

    get = resource_types[0]
    post = resource_types[1]

    assert get.name == "base"
    assert post.name == "base"
    assert len(get.uri_params) == 1
    assert len(post.uri_params) == 1

    assert get.uri_params[0].name == "mediaTypeExtension"
    assert get.uri_params[0].enum == [".json"]
    desc = "Use .json to specify application/json media type."
    assert get.uri_params[0].description.raw == desc

    assert post.uri_params[0].name == "mediaTypeExtension"
    assert post.uri_params[0].enum == [".json"]
    desc = "Use .json to specify application/json media type."
    assert post.uri_params[0].description.raw == desc

    assert len(get.responses) == 15
    assert len(post.responses) == 15

    resp = get.responses
    assert resp[0].code == 200
    assert resp[0].description.raw == "Success!"
    assert resp[1].code == 304
    assert resp[1].description.raw == "There was no new data to return."
    assert resp[2].code == 400
    desc = ("The request was invalid or cannot be otherwise served. An "
            "accompanying\nerror message will explain further. In API v1.1, "
            "requests withou\nauthentication are considered invalid and "
            "will yield this response.\n")
    assert resp[2].description.raw == desc
    assert resp[3].code == 401
    desc = "Authentication credentials were missing or incorrect."
    assert resp[3].description.raw == desc
    assert resp[4].code == 403
    desc = ("The request is understood, but it has been refused or access "
            "is no\nallowed. An accompanying error message will explain "
            "why. This code is\nused when requests are being denied due to "
            "update limits.\n")
    assert resp[5].code == 404
    desc = ("The URI requested is invalid or the resource requested, such as "
            "a user,\ndoes not exists. Also returned when the requested "
            "format is not supported\nby the requested method.\n")
    assert resp[5].description.raw == desc
    assert resp[6].code == 406
    desc = ("Returned by the Search API when an invalid format is specified "
            "in the\nrequest.\n")
    assert resp[6].description.raw == desc
    assert resp[7].code == 410
    desc = ("This resource is gone. Used to indicate that an API endpoint has "
            "been\nturned off. For example: \"The Twitter REST API v1 will "
            "soon stop\nfunctioning. Please migrate to API v1.1.\"\n")
    assert resp[7].description.raw == desc
    assert resp[8].code == 420
    desc = ("Returned by the version 1 Search and Trends APIs when you are "
            "being rate\nlimited.\n")
    assert resp[8].description.raw == desc
    assert resp[9].code == 422
    desc = ("Returned when an image uploaded to POST account/"
            "update_profile_banner is\nunable to be processed.\n")
    assert resp[9].description.raw == desc
    assert resp[10].code == 429
    desc = ("Returned in API v1.1 when a request cannot be served due to the\n"
            "application's rate limit having been exhausted for the resource. "
            "See Rate\nLimiting in API v1.1.\n")
    assert resp[10].description.raw == desc
    assert resp[11].code == 500
    desc = ("Something is broken. Please post to the group so the Twitter "
            "team can\ninvestigate.\n")
    assert resp[11].description.raw == desc
    assert resp[12].code == 502
    desc = ("Twitter is down or being upgraded.")
    assert resp[12].description.raw == desc
    assert resp[13].code == 503
    desc = ("The Twitter servers are up, but overloaded with requests. "
            "Try again later.\n")
    assert resp[13].description.raw == desc
    assert resp[14].code == 504
    desc = ("The Twitter servers are up, but the request couldn't be serviced "
            "due to\nsome failure within our stack. Try again later.\n")
    assert resp[14].description.raw == desc

    resp = post.responses
    assert resp[0].code == 200
    assert resp[0].description.raw == "Success!"
    assert resp[1].code == 304
    assert resp[1].description.raw == "There was no new data to return."
    assert resp[2].code == 400
    desc = ("The request was invalid or cannot be otherwise served. An "
            "accompanying\nerror message will explain further. In API v1.1, "
            "requests withou\nauthentication are considered invalid and "
            "will yield this response.\n")
    assert resp[2].description.raw == desc
    assert resp[3].code == 401
    desc = "Authentication credentials were missing or incorrect."
    assert resp[3].description.raw == desc
    assert resp[4].code == 403
    desc = ("The request is understood, but it has been refused or access "
            "is no\nallowed. An accompanying error message will explain "
            "why. This code is\nused when requests are being denied due to "
            "update limits.\n")
    assert resp[5].code == 404
    desc = ("The URI requested is invalid or the resource requested, such as "
            "a user,\ndoes not exists. Also returned when the requested "
            "format is not supported\nby the requested method.\n")
    assert resp[5].description.raw == desc
    assert resp[6].code == 406
    desc = ("Returned by the Search API when an invalid format is specified "
            "in the\nrequest.\n")
    assert resp[6].description.raw == desc
    assert resp[7].code == 410
    desc = ("This resource is gone. Used to indicate that an API endpoint has "
            "been\nturned off. For example: \"The Twitter REST API v1 will "
            "soon stop\nfunctioning. Please migrate to API v1.1.\"\n")
    assert resp[7].description.raw == desc
    assert resp[8].code == 420
    desc = ("Returned by the version 1 Search and Trends APIs when you are "
            "being rate\nlimited.\n")
    assert resp[8].description.raw == desc
    assert resp[9].code == 422
    desc = ("Returned when an image uploaded to POST account/"
            "update_profile_banner is\nunable to be processed.\n")
    assert resp[9].description.raw == desc
    assert resp[10].code == 429
    desc = ("Returned in API v1.1 when a request cannot be served due to the\n"
            "application's rate limit having been exhausted for the resource. "
            "See Rate\nLimiting in API v1.1.\n")
    assert resp[10].description.raw == desc
    assert resp[11].code == 500
    desc = ("Something is broken. Please post to the group so the Twitter "
            "team can\ninvestigate.\n")
    assert resp[11].description.raw == desc
    assert resp[12].code == 502
    desc = ("Twitter is down or being upgraded.")
    assert resp[12].description.raw == desc
    assert resp[13].code == 503
    desc = ("The Twitter servers are up, but overloaded with requests. "
            "Try again later.\n")
    assert resp[13].description.raw == desc
    assert resp[14].code == 504
    desc = ("The Twitter servers are up, but the request couldn't be serviced "
            "due to\nsome failure within our stack. Try again later.\n")
    assert resp[14].description.raw == desc

    assert get.headers is None
    assert post.headers is None
    assert get.body is None
    assert post.body is None
    assert get.query_params is None
    assert post.query_params is None
    assert get.form_params is None
    assert post.form_params is None
    assert get.method == "get"
    assert post.method == "post"


@pytest.fixture
def traits(api):
    return api.traits


def test_traits(traits):
    assert len(traits) == 2
    for t in traits:
        assert isinstance(t, TraitNode)

    nest = traits[0]
    assert nest.name == "nestedable"
    assert len(nest.query_params) == 1
    assert nest.query_params[0].name == "include_entities"
    assert nest.query_params[0].enum == [0, 1, True, False, None, "f"]
    desc = "The entities node will not be included when set to false."
    assert nest.query_params[0].description.raw == desc
    assert nest.form_params is None
    assert nest.uri_params is None
    assert nest.headers is None
    assert nest.responses is None
    assert nest.body is None

    trim = traits[1]
    assert trim.name == "trimmable"
    assert len(trim.query_params) == 1
    assert trim.query_params[0].name == "trim_user"
    assert trim.query_params[0].enum == [0, 1, True, False, None, "f"]
    desc = ("When set to either true, t or 1, each tweet returned in a "
            "timeline will\ninclude a user object including only the status "
            "authors numerical ID.\nOmit this parameter to receive the "
            "complete user object.\n")
    assert trim.query_params[0].description.raw == desc
    assert trim.form_params is None
    assert trim.uri_params is None
    assert trim.headers is None
    assert trim.body is None
    assert trim.responses is None


@pytest.fixture
def resources(api):
    return api.resources


ATTRIBUTES = [
    "absolute_uri", "base_uri_params", "body", "description", "display_name",
    "form_params", "headers", "is_", "media_type", "method", "name", "parent",
    "path", "protocols", "query_params", "raw", "resource_type", "responses",
    "root", "secured_by", "security_schemes", "traits", "type", "uri_params"
]

STR_ATTRS = [
    "absolute_uri", "display_name", "media_type", "method", "name",
    "path", "type"
]

ABSOLUTE_URI = "https://api.twitter.com/1.1/"


def test_resources(resources):
    assert len(resources) == 109
    for r in resources:
        assert isinstance(r, ResourceNode)


def test_resources_statuses(resources):
    res = resources[0]

    assert res.method is None
    assert res.name == "/statuses"
    assert res.display_name == "/statuses"
    assert res.type is None
    assert res.traits is None
    assert res.is_ is None
    assert res.resource_type is None
    assert res.description.raw is None
    assert res.headers is None
    assert res.body is None
    assert res.responses is None
    assert res.uri_params is None
    assert res.query_params is None
    assert res.form_params is None
    assert res.parent is None
    assert res.media_type is None
    assert res.absolute_uri == "https://api.twitter.com/1.1/statuses"
    assert res.path == "/statuses"
    assert res.protocols == ["HTTPS"]
    assert res.secured_by == ["oauth_1_0"]
    assert len(res.security_schemes) == 1


def test_resources_statuses_mention_timeline(resources):
    res = resources[1]

    desc = ("Returns the 20 most recent mentions (tweets containing a users's "
            "@screen_name)\nfor the authenticating user.\nThe timeline "
            "returned is the equivalent of the one seen when you view your\n"
            "mentions on twitter.com.\nThis method can only return up to "
            "800 tweets.\n")

    exp_data = {
        "absolute_uri": ABSOLUTE_URI + "statuses/mentions_timeline{mediaTypeExtension}",  # NOQA
        "display_name": "/mentions_timeline{mediaTypeExtension}",
        "media_type": "application/json",
        "method": "get",
        "name": "/mentions_timeline{mediaTypeExtension}",
        "path": "/statuses/mentions_timeline{mediaTypeExtension}",
        "type": "base"
    }
    for key, value in exp_data.items():
        data = getattr(res, key)
        assert value == data

    assert res.description.raw == desc
    assert res.secured_by == ["oauth_1_0"]
    # actual sec scheme objects may be different since resource could
    # have certain scopes assigned.
    assert res.security_schemes[0].raw == res.root.security_schemes[0].raw
    assert res.protocols == ["HTTPS"]
    assert res.headers is None
    assert res.body is None
    assert res.form_params is None
    assert res.is_ == ["nestedable", "trimmable"]
    assert res.type == "base"

    assert len(res.traits) == 2
    assert res.traits[0] == res.root.traits[0]
    assert res.traits[1] == res.root.traits[1]

    assert res.resource_type == res.root.resource_types[0]

    assert len(res.uri_params) == 1
    assert res.uri_params[0].name == "mediaTypeExtension"
    assert res.uri_params[0].enum == [".json"]
    desc = "Use .json to specify application/json media type."
    assert res.uri_params[0].description.raw == desc

    assert len(res.query_params) == 6

    param = res.query_params[0]
    assert param.name == "count"
    assert param.type == "integer"
    assert param.maximum == 200
    desc = ("Specifies the number of tweets to try and retrieve, up to a "
            "maximum of\n200. The value of count is best thought of as a "
            "limit to the number of\ntweets to return because suspended or "
            "deleted content is removed after\nthe count has been applied. "
            "We include retweets in the count, even if\ninclude_rts is not "
            "supplied. It is recommended you always send include_rts=1\n"
            "when using this API method.\n")
    assert param.description.raw == desc

    param = res.query_params[1]
    assert param.name == "since_id"
    assert param.type == "integer"
    desc = ("Returns results with an ID greater than (that is, more recent "
            "than) the\nspecified ID. There are limits to the number of "
            "Tweets which can be accessed\nthrough the API. If the limit of "
            "Tweets has occured since the since_id, the\nsince_id will be "
            "forced to the oldest ID available.\n")
    assert param.description.raw == desc

    param = res.query_params[2]
    assert param.name == "max_id"
    assert param.type == "integer"
    desc = ("Returns results with an ID less than (that is, older than) or "
            "equal to\nthe specified ID.\n")
    assert param.description.raw == desc

    param = res.query_params[3]
    assert param.name == "contributor_details"
    assert param.type == "string"
    desc = ("This parameter enhances the contributors element of the status "
            "response\nto include the screen_name of the contributor. By "
            "default only the user_id\nof the contributor is included.\n")
    assert param.description.raw == desc

    param = res.query_params[4]
    assert param.name == "include_entities"
    assert param.enum == [0, 1, True, False, None, "f"]
    desc = "The entities node will not be included when set to false."
    assert param.description.raw == desc

    param = res.query_params[5]
    assert param.name == "trim_user"
    assert param.enum == [0, 1, True, False, None, "f"]
    desc = ("When set to either true, t or 1, each tweet returned in a "
            "timeline will\ninclude a user object including only the status "
            "authors numerical ID.\nOmit this parameter to receive the "
            "complete user object.\n")
    assert param.description.raw == desc

    assert len(res.responses) == 15

    r = res.responses[0]

    assert r.code == 200
    assert r.description.raw == "Success!"
    assert len(r.body) == 1
    assert r.body[0].schema is not None
    assert r.body[0].example is not None
    assert r.headers is None
    assert r.config is not None

    r = res.responses[1]

    assert r.code == 304
    desc = "There was no new data to return."
    assert r.description.raw == desc
    assert r.body is None
    assert r.headers is None
    assert r.config is not None

    r = res.responses[2]
    assert r.code == 400
    desc = ("The request was invalid or cannot be otherwise served. An "
            "accompanying\nerror message will explain further. In API v1.1, "
            "requests withou\nauthentication are considered invalid and will "
            "yield this response.\n")
    assert r.description.raw == desc
    assert r.body is None
    assert r.headers is None
    assert r.config is not None

    r = res.responses[3]
    assert r.code == 401
    desc = ("Authentication credentials were missing or incorrect.")
    assert r.description.raw == desc
    assert r.body is None
    assert r.headers is None
    assert r.config is not None

    r = res.responses[4]
    assert r.code == 403
    desc = ("The request is understood, but it has been refused or access "
            "is no\nallowed. An accompanying error message will explain why. "
            "This code is\nused when requests are being denied due to "
            "update limits.\n")
    assert r.description.raw == desc
    assert r.body is None
    assert r.headers is None
    assert r.config is not None

    r = res.responses[5]
    assert r.code == 404
    desc = ("The URI requested is invalid or the resource requested, such as "
            "a user,\ndoes not exists. Also returned when the requested "
            "format is not supported\nby the requested method.\n")
    assert r.description.raw == desc
    assert r.body is None
    assert r.headers is None
    assert r.config is not None

    r = res.responses[6]
    assert r.code == 406
    desc = ("Returned by the Search API when an invalid format is specified "
            "in the\nrequest.\n")
    assert r.description.raw == desc
    assert r.body is None
    assert r.headers is None
    assert r.config is not None

    r = res.responses[7]
    assert r.code == 410
    desc = ("This resource is gone. Used to indicate that an API endpoint "
            "has been\nturned off. For example: \"The Twitter REST API "
            "v1 will soon stop\nfunctioning. Please migrate to API v1.1.\"\n")
    assert r.description.raw == desc
    assert r.body is None
    assert r.headers is None
    assert r.config is not None

    r = res.responses[8]
    assert r.code == 420
    desc = ("Returned by the version 1 Search and Trends APIs when you are "
            "being rate\nlimited.\n")
    assert r.description.raw == desc
    assert r.body is None
    assert r.headers is None
    assert r.config is not None

    r = res.responses[9]
    assert r.code == 422
    desc = ("Returned when an image uploaded to POST account/update_profile"
            "_banner is\nunable to be processed.\n")
    assert r.description.raw == desc
    assert r.body is None
    assert r.headers is None
    assert r.config is not None

    r = res.responses[10]
    assert r.code == 429
    desc = ("Returned in API v1.1 when a request cannot be served due to the\n"
            "application's rate limit having been exhausted for the resource. "
            "See Rate\nLimiting in API v1.1.\n")
    assert r.description.raw == desc
    assert r.body is None
    assert r.headers is None
    assert r.config is not None

    r = res.responses[11]
    assert r.code == 500
    desc = ("Something is broken. Please post to the group so the Twitter "
            "team can\ninvestigate.\n")
    assert r.description.raw == desc
    assert r.body is None
    assert r.headers is None
    assert r.config is not None

    r = res.responses[12]
    assert r.code == 502
    desc = "Twitter is down or being upgraded."
    assert r.description.raw == desc
    assert r.body is None
    assert r.headers is None
    assert r.config is not None

    r = res.responses[13]
    assert r.code == 503
    desc = ("The Twitter servers are up, but overloaded with requests. Try "
            "again later.\n")
    assert r.description.raw == desc
    assert r.body is None
    assert r.headers is None
    assert r.config is not None

    r = res.responses[14]
    assert r.code == 504
    desc = ("The Twitter servers are up, but the request couldn't be "
            "serviced due to\nsome failure within our stack. Try again "
            "later.\n")
    assert r.description.raw == desc
    assert r.body is None
    assert r.headers is None
    assert r.config is not None


def test_resources_statuses_user_timeline(resources):
    res = resources[2]

    desc = ("Returns a collection of the most recent Tweets posted by the "
            "user indicated\nby the screen_name or user_id parameters.\n"
            "User timelines belonging to protected users may only be "
            "requested when the\nauthenticated user either \"owns\" the "
            "timeline or is an approved follower of\nthe owner.\n"
            "The timeline returned is the equivalent of the one seen when "
            "you view a user's\nprofile on twitter.com.\nThis method can only "
            "return up to 3,200 of a user's most recent Tweets. Native\n"
            "retweets of other statuses by the user is included in this "
            "total, regardless\nof whether include_rts is set to false "
            "when requesting this resource.\n")
    assert res.description.raw == desc
